# -*- coding:utf-8 -*-

import os, platform, datetime

__project__ = "pigit"
__version__ = "1.6.0"
__url__ = "https://github.com/zlj-zz/pigit.git"
__uri__ = __url__

__author__ = "Zachary Zhang"
__email__ = "zlj19971222@outlook.com"

__license__ = "MIT"
__copyright__ = f"Copyright (c) 2021-{datetime.datetime.now().year} Zachary"

# =========================================================
# Part of compatibility.
# Handled the incompatibility between python2 and python3.
# =========================================================
VERSION = __version__

IS_WIN: bool = platform.system() == "Windows"

# For windows should have a different home path.
USER_HOME: str = ""
PIGIT_HOME: str = ""
if IS_WIN:
    USER_HOME = os.environ["USERPROFILE"]
    PIGIT_HOME = os.path.join(USER_HOME, __project__)
else:
    # ~/.config/pigit
    USER_HOME = os.environ["HOME"]
    PIGIT_HOME = os.path.join(USER_HOME, ".config", __project__)

LOG_FILE_PATH: str = f"{PIGIT_HOME}/log/{__project__}.log"

CONFIG_FILE_PATH: str = f"{PIGIT_HOME}/pigit.conf"

COUNTER_DIR_PATH: str = f"{PIGIT_HOME}/Counter"

EXTRA_CMD_MODULE_NAME: str = "extra_cmds"
EXTRA_CMD_MODULE_PATH: str = f"{PIGIT_HOME}/{EXTRA_CMD_MODULE_NAME}.py"

REPOS_PATH: str = f"{PIGIT_HOME}/repos.json"

IS_FIRST_RUN: bool = not os.path.isdir(PIGIT_HOME)
if IS_FIRST_RUN:
    os.makedirs(PIGIT_HOME, exist_ok=True)



# -*- coding:utf-8 -*-
from shutil import get_terminal_size

from .console import Term


class Screen(object):
    """
    Params:
        widget (Widget): main widget be needed by screen.
        alt_screen (bool): whether switch to a new screen.

    Attributes:
        _size (tuple): the terminal can be using size (width, height).
    """

    def __init__(self, widget=None, alt_screen: bool = True):
        self._widget = widget
        self._alt_screen = alt_screen

        self._size = None

    def start(self):
        if self._alt_screen:
            print(Term.alt_screen + Term.hide_cursor, end="")
        self.render()  # first render.

    def stop(self):
        if self._alt_screen:
            print(Term.normal_screen + Term.show_cursor, end="")

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.stop()

    def resize(self):
        """When the size has changed, this method will be call by `.loop.Loop` and try to render again."""
        self._size = None
        self.render()

    def process_input(self, key: str):
        self._widget._process_event(key)
        self.render()

    def process_mouse(self, mouse):
        # TODO: need finish.
        pass

    def render(self, resize: bool = False):
        if not self._size:
            self._size = get_terminal_size()

        print(Term.clear_screen)  # Clear full screen and cursor goto top.
        self._widget._render(self._size)



# -*- coding:utf-8 -*-
from typing import Optional
import os, time
import pytest

from pigit.codecounter import CodeCounter


@pytest.mark.parametrize(
    "path",
    [
        None,
        os.getcwd(),
        # os.environ["HOME"] + "/.config",
        # *os.environ.values(),
    ],
)
def test_pure_walk(path: Optional[str]):
    print(f"\n===Test pure walk ({path})===")
    start_t = time.time()
    cc = CodeCounter(
        count_path=path,
        format_type="table",
        whether_save=False,
    )
    cc.run()
    cc.run(format_type="simple")
    print(f"Result spend time: {time.time() - start_t}")



import os, sys, pytest
from unittest.mock import patch, Mock

from .conftest import TEST_PATH
from pigit.common.utils import exec_cmd
from pigit.gitlib.options import GitOption
from pigit.gitlib.processor import ShortGitter, get_extra_cmds
from pigit.gitlib._cmd_func import add, set_email_and_username, fetch_remote_branch
from pigit.gitlib.ignore import get_ignore_source, create_gitignore, IGNORE_TEMPLATE


class TestGitOption:
    git = GitOption()
    if not git.git_version:
        exit(1)

    # =================
    # create test repo
    # =================
    test_repo = os.path.join(TEST_PATH, "test_repo")
    os.makedirs(test_repo, exist_ok=True)
    print(exec_cmd("git init", cwd=test_repo))
    with open(os.path.join(test_repo, "test1.txt"), "w") as f:
        f.write("""This is a test file.""")
    print(exec_cmd("git add .", cwd=test_repo))
    print(exec_cmd("git commit -m 'init'", cwd=test_repo))
    with open(os.path.join(test_repo, "test2.py"), "w") as f:
        f.write("""print('This is test py file.')""")

    git.update_setting(
        op_path=test_repo, repo_info_path=os.path.join(test_repo, "reps.json")
    )

    def test(self):
        git = self.git
        print()

        print(git.get_head())
        print(git.get_branches())
        print(git.get_first_pushed_commit())
        print(git.get_remotes())
        print(git.get_remote_url())
        print(git.get_repo_desc())

    def test1(self):
        git = self.git
        print(git.load_branches())
        print(git.load_log())
        print(git.load_status())
        print(git.load_file_diff("example.py", tracked=False))
        print(git.load_commits("main"))
        print(git.load_commit_info(git.get_first_pushed_commit()))

    @pytest.mark.parametrize(
        ["get_path", "expected"],
        [
            [("err", ""), ("", "")],
            [("", "a/b/.git"), ("a/b", "a/b/.git")],
            [("", "a/b/.git/modules/"), ("a/b/", "a/b/.git/modules/")],
        ],
    )
    @patch("pigit.gitlib.options.exec_cmd")
    def test_get_repo_info(self, mock_exec_cmd, get_path, expected):
        mock_exec_cmd.return_value = get_path
        assert self.git.get_repo_info() == expected

    def test_get_repo_info_2(self):
        git = self.git
        test_repo = self.test_repo
        assert git.get_repo_info() == (test_repo, os.path.join(test_repo, ".git"))
        assert git.get_repo_info("xxxxxxx") == ("", "")

    # def pytest_sessionfinish(session, exitstatus):
    #     """whole test run finishes."""
    #     print("finish")
    #     os.rmdir(test_repo)


class TestShortGitter:
    def test_init_error(self):
        with pytest.raises(TypeError):
            ShortGitter(extra_cmds="xxx")

    @pytest.fixture(scope="module")
    def setup(self):
        extra = {"aa": {"help": "print system user name."}}
        return ShortGitter(extra_cmds=extra)

    @pytest.mark.parametrize(
        "command",
        [
            "git status",
            "git add xxx/xxx",
            "git checkout -b test",
            "git log --online --graph",
            "git log --dep 10 --online --graph",
            "git log --dep 10 --online --graph --color true",
            'git log --topo-order --stat --pretty=format:"%C(bold yellow)commit"',
        ],
    )
    def test_color_command(self, setup, command: str):
        from plenty import get_console

        console = get_console()
        handle = setup

        color_str = handle.color_command(command)
        console.echo(color_str)

    def test_load_extra_cmds(self):
        """Test load extra custom cmds."""

        name = "test_module"
        file = f"./{name}.py"
        with open(file, "w") as f:
            f.write("""extra_cmds = { 'A': 1 }""")

        d = get_extra_cmds(name, file)
        assert d["A"] == 1

        os.remove(file)


class TestCmdFunc:
    @patch("pigit.gitlib._cmd_func.exec_cmd", return_value=None)
    def test_add(self, _):
        add([])

    @patch("pigit.gitlib._cmd_func.exec_cmd", return_value=None)
    def test_fetch_remote(self, _):
        fetch_remote_branch([])

    @pytest.mark.parametrize(
        "args",
        [
            (),
            ("--global",),
            ("global",),
            ("-g",),
        ],
    )
    @patch("builtins.input", return_value="abc@gmail.com")
    @patch("pigit.gitlib._cmd_func.exec_cmd", return_value=False)
    def test_set_ua(self, _a, _b, args):
        set_email_and_username(args)


def test_iter_ignore():
    for t in IGNORE_TEMPLATE:
        print(get_ignore_source(t))


@pytest.mark.parametrize(
    ["type_", "file_name", "dir_path", "writting"],
    [
        ["xxxxxx", "ignore_text", TEST_PATH, False],
        ["rust", "ignore_test", TEST_PATH, False],
        ["rust", "ignore_test", TEST_PATH, True],
    ],
)
def test_ignore(type_, file_name, dir_path, writting):
    code, msg = create_gitignore(
        type_, file_name=file_name, dir_path=dir_path, writting=writting
    )
    print(code, msg)



# -*- coding:utf-8 -*-

from typing import Dict
import os
import textwrap

from .base import ShellCompletion


class FishCompletion(ShellCompletion):

    _SHELL: str = "fish"

    try:
        _INJECT_PATH: str = os.environ["HOME"] + "/.config/fish/config.fish"
    except:
        _INJECT_PATH: str = ""

    _template_source: str = textwrap.dedent(
        """\
        function %(func_name)s;
            set -l response;

            for value in (env %(complete_vars)s=fish_complete COMP_WORDS=(commandline -cp) \
        COMP_CWORD=(commandline -t) %(prop)s);
                set response $response $value;
            end;

            for completion in $response;
                set -l metadata (string split "," $completion);

                if test $metadata[1] = "dir";
                    __fish_complete_directories $metadata[2];
                else if test $metadata[1] = "file";
                    __fish_complete_path $metadata[2];
                else if test $metadata[1] = "plain";
                    print $metadata[2];
                end;
            end;
        end;

        complete --no-files --command %(prop)s --arguments \
        "(%(func_name)s)";
        """
    )

    def generate(self) -> str:
        # TODO:improve `fish` completion script.
        comp_keys = []

        _arguments, _positions, _sub_opts = self._parse(self.complete_vars["args"])

        for x in _arguments:
            comp_keys.extend(x[0].split())

        sub_q = [_sub_opts]
        while sub_q:
            temp: Dict = sub_q.pop(0)
            for opt_name, p in temp.items():
                comp_keys.append(opt_name)

                if _a := p.get("_arguments"):
                    for x in _a:
                        comp_keys.extend(x[0].split())
                if _opts := p.get("_sub_opts"):
                    sub_q.insert(-1, _opts)

        return " ".join(comp_keys)



# -*- coding:utf-8 -*-

"""Git optional short command dictionary.

The key is a short command, and the value is the information dictionary of the
short command.

[belong] is the type of command. The default is `Extra`.
[command] is the execution content of the short command. It supports (method, command
          string). The default is command.
[help] is help information, optional.
[has_arguments] indicates whether the short command receives parameters. The default is False.
"""

import enum
from ._cmd_func import *


@enum.unique
class CommandType(enum.Enum):
    Branch = "Branch"
    Commit = "Commit"
    Conflict = "Conflict"
    Fetch = "Fetch"
    Index = "Index"
    Log = "Log"
    Merge = "Merge"
    Push = "Push"
    Remote = "Remote"
    Stash = "Stash"
    Tag = "Tag"
    WorkingTree = "Working tree"
    Submodule = "Submodule"
    Setting = "Setting"
    Extra = "Extra"  # default


# The custom git output format string.
#   git ... --pretty={0}.format(GIT_PRINT_FORMAT)
_GIT_PRINT_FORMAT = (
    'format:"%C(bold yellow)commit %H%C(auto)%d%n'
    "%C(bold)Author: %C(blue)%an <%ae> %C(reset)%C(cyan)%ai (%ar)%n"
    '%C(bold)Commit: %C(blue)%cn <%ce> %C(reset)%C(cyan)%ci (%cr)%C(reset)%n%+B"'
)


GIT_CMDS = {
    # Branch
    "b": {
        "belong": CommandType.Branch,
        "command": "git branch",
        "help": "lists, creates, renames, and deletes branches.",
        "has_arguments": True,
    },
    "bc": {
        "belong": CommandType.Branch,
        "command": "git checkout -b",
        "help": "creates a new branch.",
        "has_arguments": True,
    },
    "bl": {
        "belong": CommandType.Branch,
        "command": "git branch -vv",
        "help": "lists branches and their commits.",
    },
    "bL": {
        "belong": CommandType.Branch,
        "command": "git branch --all -vv",
        "help": "lists local and remote branches and their commits.",
    },
    "bs": {
        "belong": CommandType.Branch,
        "command": "git show-branch",
        "help": "lists branches and their commits with ancestry graphs.",
    },
    "bS": {
        "belong": CommandType.Branch,
        "command": "git show-branch --all",
        "help": "lists local and remote branches and their commits with "
        "ancestry graphs.",
    },
    "bm": {
        "belong": CommandType.Branch,
        "command": "git branch --move",
        "help": "renames a branch.",
        "has_arguments": True,
    },
    "bM": {
        "belong": CommandType.Branch,
        "command": "git branch --move --force",
        "help": "renames a branch even if the new branch name already exists.",
        "has_arguments": True,
    },
    "bd": {
        "belong": CommandType.Branch,
        "command": "git branch -d",
        "help": "delete a local branch by name.",
        "has_arguments": True,
    },
    # Commit
    "c": {
        "belong": CommandType.Commit,
        "command": "git commit --verbose",
        "help": "records changes to the repository.",
    },
    "ca": {
        "belong": CommandType.Commit,
        "command": "git commit --verbose --all",
        "help": "commits all modified and deleted files.",
    },
    "cA": {
        "belong": CommandType.Commit,
        "command": "git commit --verbose --patch",
        "help": "commits all modified and deleted files interactively",
    },
    "cm": {
        "belong": CommandType.Commit,
        "command": "git commit --verbose --message",
        "help": "commits with the given message.",
    },
    "co": {
        "belong": CommandType.Commit,
        "command": "git checkout",
        "help": "checks out a branch or paths to the working tree.",
        "has_arguments": True,
    },
    "cO": {
        "belong": CommandType.Commit,
        "command": "git checkout --patch",
        "help": "checks out hunks from the index or the tree interactively.",
        "has_arguments": True,
    },
    "cf": {
        "belong": CommandType.Commit,
        "command": "git commit --amend --reuse-message HEAD ",
        "help": "amends the tip of the current branch reusing the same log "
        "message as HEAD.",
    },
    "cF": {
        "belong": CommandType.Commit,
        "command": "git commit --verbose --amend",
        "help": "amends the tip of the current branch.",
    },
    "cr": {
        "belong": CommandType.Commit,
        "command": "git revert",
        "help": "reverts existing commits by reverting patches and recording "
        "new commits.",
        "has_arguments": True,
    },
    "cR": {
        "belong": CommandType.Commit,
        "command": 'git reset "HEAD^"',
        "help": "removes the HEAD commit.",
    },
    "cs": {
        "belong": CommandType.Commit,
        "command": f"git show --pretty={_GIT_PRINT_FORMAT}",
        "help": "shows one or more objects (blobs, trees, tags and commits).",
        "has_arguments": True,
    },
    # Conflict(C)
    "Cl": {
        "belong": CommandType.Conflict,
        "command": "git --no-pager diff --diff-filter=U --name-only",
        "help": "lists unmerged files.",
    },
    "Ca": {
        "belong": CommandType.Conflict,
        "command": "git add git --no-pager diff --diff-filter=U --name-only",
        "help": "adds unmerged file contents to the index.",
        "has_arguments": True,
    },
    "Ce": {
        "belong": CommandType.Conflict,
        "command": "git mergetool git --no-pager diff --diff-filter=U --name-only",
        "help": "executes merge-tool on all unmerged files.",
    },
    "Co": {
        "belong": CommandType.Conflict,
        "command": "git checkout --ours -- ",
        "help": "checks out our changes for unmerged paths.",
    },
    "CO": {
        "belong": CommandType.Conflict,
        "command": "git checkout --ours -- git --no-pager diff --diff-filter=U --name-only",
        "help": "checks out our changes for all unmerged paths.",
    },
    "Ct": {
        "belong": CommandType.Conflict,
        "command": "git checkout --theirs -- ",
        "help": "checks out their changes for unmerged paths.",
    },
    "CT": {
        "belong": CommandType.Conflict,
        "command": "git checkout --theirs -- git --no-pager diff --diff-filter=U --name-only",
        "help": "checks out their changes for all unmerged paths.",
    },
    # Fetch(f)
    "f": {
        "belong": CommandType.Fetch,
        "command": "git fetch",
        "help": "downloads objects and references from another repository.",
        "has_arguments": True,
    },
    "fc": {
        "belong": CommandType.Fetch,
        "command": "git clone",
        "help": "clones a repository into a new directory.",
        "has_arguments": True,
    },
    "fC": {
        "belong": CommandType.Fetch,
        "command": "git clone --depth=1",
        "help": "clones a repository into a new directory clearly(depth:1).",
        "has_arguments": True,
    },
    "fm": {
        "belong": CommandType.Fetch,
        "command": "git pull",
        "help": "fetches from and merges with another repository or local branch.",
        "has_arguments": True,
    },
    "fr": {
        "belong": CommandType.Fetch,
        "command": "git pull --rebase",
        "help": "fetches from and rebase on top of another repository or local branch.",
        "has_arguments": True,
    },
    "fu": {
        "belong": CommandType.Fetch,
        "command": "git fetch --all --prune && git merge --ff-only @{u}",
        "help": "removes un-existing remote-tracking references, fetches all remotes "
        "and merges.",
        "has_arguments": True,
    },
    "fb": {
        "belong": CommandType.Fetch,
        "command": fetch_remote_branch,
        "help": "fetch other branch to local as same name.",
    },
    # Index(i)
    "ia": {
        "belong": CommandType.Index,
        "command": add,
        "help": "adds file contents to the index(default: all files).",
        "has_arguments": True,
    },
    "iA": {
        "belong": CommandType.Index,
        "command": "git add --patch",
        "help": "adds file contents to the index interactively.",
        "has_arguments": True,
    },
    "iu": {
        "belong": CommandType.Index,
        "command": "git add --update",
        "help": "adds file contents to the index (updates only known files).",
        "has_arguments": True,
    },
    "id": {
        "belong": CommandType.Index,
        "command": "git diff --no-ext-diff --cached",
        "help": "displays changes between the index and a named commit (diff).",
        "has_arguments": True,
    },
    "iD": {
        "belong": CommandType.Index,
        "command": "git diff --no-ext-diff --cached --word-diff",
        "help": "displays changes between the index and a named commit (word diff).",
        "has_arguments": True,
    },
    "ir": {
        "belong": CommandType.Index,
        "command": "git reset",
        "help": "resets the current HEAD to the specified state.",
        "has_arguments": True,
    },
    "iR": {
        "belong": CommandType.Index,
        "command": "git reset --patch",
        "help": "resets the current index interactively.",
        "has_arguments": True,
    },
    "ix": {
        "belong": CommandType.Index,
        "command": "git rm --cached -r",
        "help": "removes files from the index (recursively).",
        "has_arguments": True,
    },
    "iX": {
        "belong": CommandType.Index,
        "command": "git rm --cached -rf",
        "help": "removes files from the index (recursively and forced).",
        "has_arguments": True,
    },
    # Log(l)
    "l": {
        "belong": CommandType.Log,
        "command": "git log --graph --all --decorate",
        "help": "display the log with good format.",
    },
    "l1": {
        "belong": CommandType.Log,
        "command": "git log --graph --all --decorate --oneline",
        "help": "display the log with oneline.",
    },
    "ls": {
        "belong": CommandType.Log,
        "command": f"git log --topo-order --stat --pretty={_GIT_PRINT_FORMAT}",
        "help": "displays the stats log.",
    },
    "ld": {
        "belong": CommandType.Log,
        "command": f"git log --topo-order --stat --patch --pretty={_GIT_PRINT_FORMAT}",
        "help": "displays the diff log.",
    },
    "lv": {
        "belong": CommandType.Log,
        "command": f"git log --topo-order --show-signature --pretty={_GIT_PRINT_FORMAT}",
        "help": "displays the log, verifying the GPG signature of commits.",
    },
    "lc": {
        "belong": CommandType.Log,
        "command": "git shortlog --summary --numbered",
        "help": "displays the commit count for each contributor in descending order.",
    },
    "lr": {
        "belong": CommandType.Log,
        "command": "git reflog",
        "help": "manages reflog information.",
        "has_arguments": True,
    },
    # Merge(m)
    "m": {
        "belong": CommandType.Merge,
        "command": "git merge",
        "help": "joins two or more development histories together.",
        "has_arguments": True,
    },
    "ma": {
        "belong": CommandType.Merge,
        "command": "git merge --abort",
        "help": "aborts the conflict resolution, and reconstructs the pre-merge state.",
        "has_arguments": True,
    },
    "mC": {
        "belong": CommandType.Merge,
        "command": "git merge --no-commit",
        "help": "performs the merge but does not commit.",
        "has_arguments": True,
    },
    "mF": {
        "belong": CommandType.Merge,
        "command": "git merge --no-ff",
        "help": "creates a merge commit even if the merge could be resolved as a "
        "fast-forward.",
        "has_arguments": True,
    },
    "mS": {
        "belong": CommandType.Merge,
        "command": "git merge -S",
        "help": "performs the merge and GPG-signs the resulting commit.",
        "has_arguments": True,
    },
    "mv": {
        "belong": CommandType.Merge,
        "command": "git merge --verify-signatures",
        "help": "verifies the GPG signature of the tip commit of the side branch "
        "being merged.",
        "has_arguments": True,
    },
    "mt": {
        "belong": CommandType.Merge,
        "command": "git mergetool",
        "help": "runs the merge conflict resolution tools to resolve conflicts.",
        "has_arguments": True,
    },
    # Push(p)
    "p": {
        "belong": CommandType.Push,
        "command": "git push",
        "help": "updates remote refs along with associated objects.",
        "has_arguments": True,
    },
    "pf": {
        "belong": CommandType.Push,
        "command": "git push --force-with-lease",
        "help": 'forces a push safely (with "lease").',
        "has_arguments": True,
    },
    "pF": {
        "belong": CommandType.Push,
        "command": "git push --force",
        "help": "forces a push.",
        "has_arguments": True,
    },
    "pa": {
        "belong": CommandType.Push,
        "command": "git push --all",
        "help": "pushes all branches.",
    },
    "pA": {
        "belong": CommandType.Push,
        "command": "git push --all && git push --tags",
        "help": "pushes all branches and tags.",
    },
    "pt": {
        "belong": CommandType.Push,
        "command": "git push --tags",
        "help": "pushes all tags.",
    },
    "pc": {
        "belong": CommandType.Push,
        "command": 'git push --set-upstream origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)"',
        "help": "pushes the current branch and adds origin as an upstream reference for it.",
    },
    "pp": {
        "belong": CommandType.Push,
        "command": (
            'git pull origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)" && '
            'git push origin "$(git symbolic-ref -q --short HEAD 2> /dev/null)"'
        ),
        "help": "pulls and pushes the current branch from origin to origin.",
    },
    # Remote(R)
    "R": {
        "belong": CommandType.Remote,
        "command": "git remote",
        "help": "manages tracked repositories.",
        "has_arguments": True,
    },
    "Rl": {
        "belong": CommandType.Remote,
        "command": "git remote --verbose",
        "help": "lists remote names and their URLs.",
        "has_arguments": True,
    },
    "Ra": {
        "belong": CommandType.Remote,
        "command": "git remote add",
        "help": "adds a new remote.",
        "has_arguments": True,
    },
    "Rx": {
        "belong": CommandType.Remote,
        "command": "git remote rm",
        "help": "removes a remote.",
        "has_arguments": True,
    },
    "Rm": {
        "belong": CommandType.Remote,
        "command": "git remote rename",
        "help": "renames a remote.",
        "has_arguments": True,
    },
    "Ru": {
        "belong": CommandType.Remote,
        "command": "git remote update",
        "help": "fetches remotes updates.",
        "has_arguments": True,
    },
    "Rp": {
        "belong": CommandType.Remote,
        "command": "git remote prune",
        "help": "prunes all stale remote tracking branches.",
        "has_arguments": True,
    },
    "Rs": {
        "belong": CommandType.Remote,
        "command": "git remote show",
        "help": "shows information about a given remote.",
        "has_arguments": True,
    },
    "RS": {
        "belong": CommandType.Remote,
        "command": "git remote set-url",
        "help": "changes URLs for a remote.",
        "has_arguments": True,
    },
    # Stash(s)
    "s": {
        "belong": CommandType.Stash,
        "command": "git stash",
        "help": "stashes the changes of the dirty working directory.",
        "has_arguments": True,
    },
    "sp": {
        "belong": CommandType.Stash,
        "command": "git stash pop",
        "help": "removes and applies a single stashed state from the stash list.",
    },
    "sl": {
        "belong": CommandType.Stash,
        "command": "git stash list",
        "help": "lists stashed states.",
    },
    "sd": {
        "belong": CommandType.Stash,
        "command": "git stash show",
        "help": "display stash list.",
        "has_arguments": True,
    },
    "sD": {
        "belong": CommandType.Stash,
        "command": "git stash show --patch --stat",
        "help": "display stash list with detail.",
        "has_arguments": True,
    },
    # Tag (t)
    "t": {
        "belong": CommandType.Tag,
        "command": "git tag",
        "help": "creates, lists, deletes or verifies a tag object signed with GPG.",
        "has_arguments": True,
    },
    "ta": {
        "belong": CommandType.Tag,
        "command": "git tag -a",
        "help": "create a new tag.",
        "has_arguments": True,
    },
    "tx": {
        "belong": CommandType.Tag,
        "command": "git tag --delete",
        "help": "deletes tags with given names.",
        "has_arguments": True,
    },
    # Working tree(w)
    "ws": {
        "belong": CommandType.WorkingTree,
        "command": "git status --short",
        "help": "displays working-tree status in the short format.",
    },
    "wS": {
        "belong": CommandType.WorkingTree,
        "command": "git status",
        "help": "displays working-tree status.",
        "has_arguments": True,
    },
    "wd": {
        "belong": CommandType.WorkingTree,
        "command": "git diff --no-ext-diff",
        "help": "displays changes between the working tree and the index (diff).",
        "has_arguments": True,
    },
    "wD": {
        "belong": CommandType.WorkingTree,
        "command": "git diff --no-ext-diff --word-diff",
        "help": "displays changes between the working tree and the index (word diff).",
        "has_arguments": True,
    },
    "wr": {
        "belong": CommandType.WorkingTree,
        "command": "git reset --soft",
        "help": "resets the current HEAD to the specified state, does not touch the "
        "index nor the working tree.",
        "has_arguments": True,
    },
    "wR": {
        "belong": CommandType.WorkingTree,
        "command": "git reset --hard",
        "help": "resets the current HEAD, index and working tree to the specified state.",
        "has_arguments": True,
    },
    "wc": {
        "belong": CommandType.WorkingTree,
        "command": "git clean --dry-run",
        "help": "cleans untracked files from the working tree (dry-run).",
        "has_arguments": True,
    },
    "wC": {
        "belong": CommandType.WorkingTree,
        "command": "git clean -d --force",
        "help": "cleans untracked files from the working tree.",
        "has_arguments": True,
    },
    "wm": {
        "belong": CommandType.WorkingTree,
        "command": "git mv",
        "help": "moves or renames files.",
        "has_arguments": True,
    },
    "wM": {
        "belong": CommandType.WorkingTree,
        "command": "git mv -f",
        "help": "moves or renames files (forced).",
        "has_arguments": True,
    },
    "wx": {
        "belong": CommandType.WorkingTree,
        "command": "git rm -r",
        "help": "removes files from the working tree and from the index (recursively).",
        "has_arguments": True,
    },
    "wX": {
        "belong": CommandType.WorkingTree,
        "command": "git rm -rf",
        "help": "removes files from the working tree and from the index (recursively "
        "and forced).",
        "has_arguments": True,
    },
    # Submodule
    "Sc": {
        "belong": CommandType.Submodule,
        "command": "git clone --recursive",
        "help": "Clone a repository as a submodule.",
        "has_arguments": True,
    },
    "Si": {
        "belong": CommandType.Submodule,
        "command": "git submodule update --init --recursive",
        "help": "Pull the submodule for the first time.",
    },
    # For above git 1.8.2
    #   `git submodule update --recursive --remote`
    # For above git 1.7.3
    #   `git submodule update --recursive`
    #   `git pull --reccurse-submodules`
    "Su": {
        "belong": CommandType.Submodule,
        "command": "git submodule update --recursive --remote",
        "help": "Update git submodule.",
    },
    "Sd": {
        "belong": CommandType.Submodule,
        "command": "git rm --cached",
        "help": "Remove submodule from repository.",
        "has_arguments": True,
    },
    "SD": {
        "belong": CommandType.Submodule,
        "command": "git submodule deinit",
        "help": "Inverse initialization submodule, clear the dir.",
        "has_arguments": True,
    },
    # Setting
    "savepd": {
        "belong": CommandType.Setting,
        "command": "git config credential.helper store",
        "help": "Remember your account and password.",
        "has_arguments": True,
    },
    "ue": {
        "belong": CommandType.Setting,
        "command": set_email_and_username,
        "help": "set email and username interactively.",
        "has_arguments": True,
    },
    "user": {
        "belong": CommandType.Setting,
        "command": "git config user.name",
        "help": "set username.",
        "has_arguments": True,
    },
    "email": {
        "belong": CommandType.Setting,
        "command": "git config user.email",
        "help": "set user email.",
        "has_arguments": True,
    },
}  # type: dict[str,dict]



# -*- coding:utf-8 -*-


class Term:
    """Terminal console."""

    hide_cursor = "\033[?25l"  # * Hide terminal cursor
    show_cursor = "\033[?25h"  # * Show terminal cursor
    alt_screen = "\033[?1049h"  # * Switch to alternate screen
    normal_screen = "\033[?1049l"  # * Switch to normal screen
    clear_screen = "\033[2J\033[0;0f"  # * Clear screen and set cursor to position 0,0

    # * Enable reporting of mouse position on click and release
    mouse_on = "\033[?1002h\033[?1015h\033[?1006h"
    mouse_off = "\033[?1002l"  # * Disable mouse reporting

    # * Enable reporting of mouse position at any movement
    mouse_direct_on = "\033[?1003h"
    mouse_direct_off = "\033[?1003l"  # * Disable direct mouse reporting


class Cursor:
    """Class with collection of cursor movement functions:
    Functions:
        .t[o](line, column)
        .r[ight](columns)
        .l[eft](columns)
        .u[p](lines)
        .d[own](lines)
        .save()
        .restore()
    """

    @staticmethod
    def to(line, col):
        # * Move cursor to line, column
        return "\033[{};{}f".format(line, col)

    @staticmethod
    def right(dx):
        return "\033[{}C".format(dx)

    @staticmethod
    def left(dx):
        return "\033[{}D".format(dx)

    @staticmethod
    def up(dy):
        return "\033[{}A".format(dy)

    @staticmethod
    def down(dy):
        return "\033[{}B".format(dy)

    save = "\033[s"  # * Save cursor position
    restore = "\033[u"  # * Restore saved cursor postion
    t = to
    r = right
    l = left
    u = up
    d = down

    @staticmethod
    def hide_cursor():
        print("\033[?25l")  # * Hide terminal cursor

    @staticmethod
    def show_cursor():
        print("\033[?25h")  # * Show terminal cursor



# -*- coding:utf-8 -*-
import pytest
from .utils import analyze_it

from pigit.info import GitConfig, introduce
from plenty import get_console

console = get_console()


def test_introduce():
    console.echo()
    console.echo(introduce())


def test_gitconfig():
    c = GitConfig()
    # c = GitConfig(format_type="normal")
    print()
    console.echo(c.generate())


@analyze_it
def test():
    introduce()

    c = GitConfig()
    c.format_type = "table"
    c.generate()
    c.format_type = "normal"
    c.generate()



# -*- coding:utf-8 -*-

from typing import Optional
import os
import logging
import logging.handlers


FMT_NORMAL = logging.Formatter(
    fmt="%(asctime)s %(levelname).4s %(message)s", datefmt="%H:%M:%S"
)

FMT_DEBUG = logging.Formatter(
    fmt="%(asctime)s.%(msecs)03d %(levelname).4s [%(name)s] %(message)s",
    datefmt="%H:%M:%S",
)


def setup_logging(debug: bool = False, log_file: Optional[str] = None):
    root_logger = logging.getLogger()
    # print(log_file)

    if debug:
        log_level = logging.DEBUG
        formatter = FMT_DEBUG
    else:
        log_level = logging.INFO
        formatter = FMT_NORMAL

    if log_file is None:
        log_handle = logging.StreamHandler()
    else:
        # try create dir of log.
        os.makedirs(os.path.dirname(log_file), exist_ok=True)

        try:
            log_handle = logging.handlers.RotatingFileHandler(
                log_file, maxBytes=1048576, backupCount=4
            )
        except PermissionError:
            print("No permission to write to '{0}' directory!".format(log_file))
            raise SystemExit(1) from None

    log_handle.setFormatter(formatter)
    log_handle.setLevel(log_level)

    root_logger.addHandler(log_handle)
    root_logger.setLevel(0)



# -*- coding:utf-8 -*-

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)
from argparse import (
    Namespace,
    ArgumentParser,
    HelpFormatter,
    _SubParsersAction,
)
from shutil import get_terminal_size
from plenty.style import Style


if TYPE_CHECKING:
    from argparse import Action, FileType, _ArgumentGroup

FormatterStyle = Union[Style, str, None]


class ColorHelpFormatter(HelpFormatter):
    """Formatter for generating usage messages and argument help strings.
    This class inherits `argparse.HelpFormatter` and rewrites some methods
    to complete customization.
    """

    usage_style: FormatterStyle = "bold"
    text_style: FormatterStyle = "i sky_blue"
    command_style: FormatterStyle = "ok"
    help_style: FormatterStyle = "i yellow"

    def __init__(
        self,
        prog: str,
        indent_increment: int = 2,
        max_help_position: int = 24,
        width: Optional[int] = None,
    ):
        max_width, _ = get_terminal_size()
        if width is None:
            width = max_width * 2 // 3
        else:
            width = width if width < max_width else max_width - 2

        super().__init__(prog, indent_increment, max_help_position, width)

    def _format_usage(
        self,
        usage: str,
        actions: Iterable["Action"],
        groups: Iterable["_ArgumentGroup"],
        prefix: Optional[str],
    ) -> str:
        return Style.parse(self.usage_style).render(
            super()._format_usage(usage, actions, groups, prefix)
        )

    def _format_text(self, text: str) -> str:
        return Style.parse(self.text_style).render(super()._format_text(text))

    def _format_action(self, action: "Action") -> str:
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2, self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)

        # no help; start on same line and add a final newline
        if not action.help:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup

        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            tup = self._current_indent, "", action_width, action_header
            action_header = "%*s%-*s  " % tup
            indent_first = 0

        # long action name; start on the next line
        else:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup
            indent_first = help_position

        # collect the pieces of the action help
        # @Overwrite
        parts = [Style.parse(self.command_style).render(action_header)]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            help_parts = ["%*s%s\n" % (indent_first, "", help_lines[0])]
            help_parts.extend(
                "%*s%s\n" % (help_position, "", line) for line in help_lines[1:]
            )
            parts.append(
                Style.parse(self.help_style).render(self._join_parts(help_parts))
            )

        elif not action_header.endswith("\n"):
            parts.append("\n")

        # if there are any sub-actions, add their help as well
        parts.extend(
            self._format_action(subaction)
            for subaction in self._iter_indented_subactions(action)
        )

        # return a single string
        return self._join_parts(parts)


class ParserError(Exception):
    """Error class of ~Parser."""


class Parser(ArgumentParser):
    def __init__(
        self,
        prog: Optional[str] = None,
        usage: Optional[str] = None,
        description: Optional[str] = None,
        epilog: Optional[str] = None,
        parents: Sequence[ArgumentParser] = None,
        formatter_class: Sequence[HelpFormatter] = ColorHelpFormatter,
        prefix_chars: str = "-",
        fromfile_prefix_chars: Optional[str] = None,
        argument_default: Any = None,
        conflict_handler: str = "error",
        add_help: bool = True,
        allow_abbrev: bool = True,
        # exit_on_error: bool = True,  # python3.9 feature
        callback: Optional[Callable] = None,
    ) -> None:
        if parents is None:
            parents = []
        super().__init__(
            prog,
            usage,
            description,
            epilog,
            parents,
            formatter_class,
            prefix_chars,
            fromfile_prefix_chars,
            argument_default,
            conflict_handler,
            add_help,
            allow_abbrev,
        )

        self.subparsers_action: Optional["_SubParsersAction"] = None
        self.default_callback = callback

    # ================================
    # overload ~ArgumentParser method
    # ================================
    def add_subparsers(self, **kwargs):
        self.subparsers_action = super().add_subparsers(**kwargs)
        return self.subparsers_action

    # ============
    # call method
    # ============
    def main(self, args: Optional[List[str]] = None, **kwargs):
        known_args: Namespace
        unknown: List

        known_args, unknown = self.parse_known_args(args)
        # print("call", known_args, unknown)

        if "sub_callback" in known_args:
            known_args.sub_callback(known_args, unknown)
        elif self.default_callback is not None:
            self.default_callback(known_args, unknown)

    def __call__(self, args: Optional[List[str]] = None, **kwds: Any) -> Any:
        self.main(args, **kwds)

    # ===============================
    # tools methods of serialization
    # ===============================
    @classmethod
    def from_dict(cls, parser_dict: Dict) -> "Parser":
        """Parse a `dict` to genrate a ~Parser."""
        from copy import deepcopy

        # Use `deepcopy` to ensure that the original dict will not be changed.
        parser_dict = deepcopy(parser_dict)

        def add_command(
            handle: Union["Parser", "_ArgumentGroup"],
            name: str,
            args: Dict[str, Any],
        ) -> None:
            """Add command to ~Parser or ~Group object."""

            names: List[str] = name.split(" ")
            handle.add_argument(*names, **args)

        def parse_args(handle: "Parser", args: Dict) -> None:
            sub_parsers: Optional["Parser"] = None

            for name, prop in args.items():
                # Get command type.
                prop_type = prop.pop("type", "")

                # If the type is 'groups', it's mean that need create a new custom
                # group. The command of the group in 'args', so we should iterative
                # the 'args' to add each command.
                if prop_type == "groups":
                    g_handle: "_ArgumentGroup" = handle.add_argument_group(
                        title=prop.get("title", ""),
                        description=prop.get("description", ""),
                    )
                    for g_name, g_prop in prop.get("args", {}).items():
                        add_command(g_handle, g_name, g_prop)

                # If the type is 'sub', is's mean that need create a new sub-parser.
                # Before create sub-parse, we need create a subparsers which a sub-parser
                # group. Be careful that cannot have multiple subparser arguments for same
                # ~ArgumentParser. The command of sub-parser in 'args' and it may include
                # smaller sub-parser, so should resolve it recursively.
                elif prop_type == "sub":
                    #
                    if not sub_parsers:
                        sub_parsers = handle.add_subparsers()

                    sub_args: Dict = prop.pop("args", None)

                    sub_handle: Sequence["Parser"] = sub_parsers.add_parser(
                        name, **prop
                    )

                    # If `set_defaults` in args, special treatment is required.
                    set_defaults: Dict = sub_args.get("set_defaults", None)
                    if set_defaults:
                        del sub_args["set_defaults"]
                        sub_handle.set_defaults(**set_defaults)

                    parse_args(sub_handle, sub_args)

                # Other types will be ignored, and it is considered that the current is
                # just an ordinary command to add.
                else:
                    add_command(handle, name, prop)

        args: Dict = parser_dict.pop("args", {})

        # Create root parser. Parse and add command.
        root_parser = cls(**parser_dict)
        parse_args(root_parser, args)

        return root_parser

    def to_dict(self) -> Dict:
        """Return a dict of a parameter serialization of ~Parser."""

        cmd_names: List[str] = [
            "prog",
            "usage",
            "description",
            "epilog",
            "prefix_chars",
        ]
        argument_names: List[str] = [
            "nargs",
            "const",
            "dest",
            "default",
            "type",
            "metavar",
            "help",
        ]

        def _process(parser: Sequence["Parser"], target_dict: Dict) -> Dict:
            # Set parser parameters.
            for name in cmd_names:
                target_dict[name] = getattr(parser, name, None)

            # Init `args`.
            target_dict["args"] = args = {}

            # Iterative action groups. Include 'option', 'position' and 'subparsers'.
            # If define custom group, also include them.
            for action_group in parser._action_groups:
                # Iterative action in the group. The each action is adding by `add_argument`,
                # so one action is one argument. It include the parameters that we needed,
                # the name is define in `argument_names`. The only special type is
                # ～_SubParserAction, which is a sub-parser. We need to deal with it
                # separately and resolve its action groups recursively.
                for action in action_group._group_actions:
                    if isinstance(action, _SubParsersAction):
                        sub_helps = {
                            choices_action.dest: choices_action.help
                            for choices_action in action._choices_actions
                        }
                        for sub_prog, sub_parser in action._name_parser_map.items():
                            args[sub_prog] = _process(sub_parser, {})
                            args[sub_prog]["help"] = sub_helps.get(sub_prog, "_")
                            args[sub_prog]["type"] = "sub"
                    else:
                        option_string = " ".join(action.option_strings)
                        args[option_string] = {}

                        for name in argument_names:
                            args[option_string][name] = getattr(action, name, None)

            return target_dict

        return _process(self, {})

    # ===============================
    # quick add sub-parser decorator
    # ===============================
    @overload
    def sub_parser(
        self,
        prog: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        parser_class: Optional["Parser"] = None,
        action: Optional["Action"] = None,
        option_string: Optional[str] = None,
        dest: Optional[str] = None,
        required: bool = False,
        help: Optional[str] = None,
        metavar: Optional[str] = None,
    ) -> "Parser":
        ...

    def sub_parser(self, prog: str, **kwargs) -> "Parser":
        """Create a new ~Parser of subparser and use the decorator use callback.
        This will also automatically attach all decorated :func:`argument` as
        parameters to the parser.

        Args:
            prog (str): the prog name of subparser.

        Raises:
            ParserError: when the prog is not str.

        Returns:
            Parser:
        """

        if not isinstance(prog, str):
            raise ParserError(
                f"The name of sub_parser must be str, but given a {type(prog).__name__}."
            ) from None

        if self.subparsers_action is None:
            self.add_subparsers()

        def decorator(fn: Callable[..., Any]) -> "Parser":
            nonlocal kwargs
            attr_params = kwargs.pop("params", None)
            params = attr_params if attr_params is not None else []

            try:
                decorator_params = fn.__parser_params__
            except AttributeError:
                pass
            else:
                del fn.__parser_params__
                params.extend(decorator_params)

            if "description" not in kwargs:
                kwargs["description"] = fn.__doc__

            parser = self.subparsers_action.add_parser(prog, **kwargs)
            for args, kwargs in params:
                parser.add_argument(*args, **kwargs)

            parser.set_defaults(sub_callback=fn)

            return parser

        return decorator


# ====================================
# decorator method of creating parser
# ====================================
@overload
def command(
    prog: Union[str, Callable, None] = None,
    cls: Sequence[Parser] = None,
    usage: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    parents: Sequence[Parser] = None,
    formatter_class: HelpFormatter = HelpFormatter,
    prefix_chars: str = "-",
    fromfile_prefix_chars: Optional[str] = None,
    argument_default: Any = None,
    conflict_handler: str = "error",
    add_help: bool = True,
    allow_abbrev: bool = True,
    exit_on_error: bool = True,
    callback: Optional[Callable] = None,
) -> Parser:
    ...


def command(
    prog: Union[str, Callable, None] = None, cls: Sequence[Parser] = None, **attrs
) -> Parser:
    """Creates a new ~Parser and uses the decorated function as callback.
    This will also automatically attach all decorated :func:`argument` as
    parameters to the parser.

    The name of the command defaults to the name of the function with
    underscores replaced by dashes. If you want to change that, you can
    pass the intended name as the first argument.
    """

    if cls is None:
        cls = Parser

    fn: Optional[Callable] = None
    if callable(prog):
        fn = prog
        prog = None

    def decorator(fn: Callable[..., Any]) -> Parser:
        kwargs = attrs
        attr_params = kwargs.pop("params", None)
        params = attr_params if attr_params is not None else []

        try:
            decorator_params = fn.__parser_params__
        except AttributeError:
            pass
        else:
            del fn.__parser_params__
            params.extend(decorator_params)

        if "description" not in kwargs:
            kwargs["description"] = fn.__doc__

        cmd = cls(
            prog=prog or fn.__name__.lower().replace("_", "-").strip(),
            callback=fn,
            **kwargs,
        )

        for args, kwargs in params:
            cmd.add_argument(*args, **kwargs)

        return cmd

    if fn is not None:
        return decorator(fn)

    return decorator


def _param_memo(fn: Callable[..., Any], params) -> None:
    if not hasattr(fn, "__parser_params__"):
        fn.__parser_params__ = []

    fn.__parser_params__.append(params)


@overload
def argument(
    name: str,
    action: Union[
        Literal[
            "store",
            "store_const",
            "store_true",
            "store_false",
            "append",
            "append_const",
            "count",
            "help",
            "version",
            "extend",
        ],
        Sequence["Action"],
        None,
    ] = None,
    nargs: Union[
        int, Literal["?", "*", "+", "...", "A...", "==SUPPRESS=="], None
    ] = None,
    const: Optional[Any] = None,
    default: Optional[Any] = None,
    type: Union[Callable, "FileType", None] = None,
    choices: Optional[Iterable] = None,
    required: Optional[bool] = None,
    help: Optional[str] = None,
    metavar: Union[str, Tuple[str, ...], None] = None,
    dest: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Any,
) -> Callable:
    ...


def argument(name: str, **kwargs) -> Callable:
    """Attaches an argument to the parser, allow position or option."""

    if not isinstance(name, str):
        raise TypeError(
            f"The type of 'name' must be str, but receive {type(name).__name__}."
        ) from None

    def decorator(fn: Callable[..., Any]) -> Callable:
        if "help" not in kwargs:
            kwargs["help"] = fn.__doc__

        # Check each name whether valid.
        names: List[str] = name.strip().split()
        _param_memo(fn, (names, kwargs))

        return fn

    return decorator



from typing import Optional, Tuple
from shutil import get_terminal_size
from pathlib import Path
from .template import IGNORE_TEMPLATE


__all__ = ("get_ignore_source", "create_gitignore")


def get_ignore_source(type_: str) -> Optional[str]:
    """Get gitignore source follow type."""
    return IGNORE_TEMPLATE.get(type_)


def create_gitignore(
    type_: str,
    file_name: str = ".gitignore",
    dir_path: Optional[str] = None,
    writting: bool = True,
) -> Tuple[int, str]:
    """Try to create a gitignore file.

    Args:
        type_ (str): project type.
        file_name (str, optional): file name. Defaults to ".gitignore".
        dir_path (Optional[str], optional): dir path. Defaults to None.
        writting (bool, optional): whether writting. Defaults to True.

    Returns:
        Tuple[int, str]: (code, message)
    """

    source = IGNORE_TEMPLATE.get(type_.lower())
    path = Path(dir_path or ".").joinpath(file_name)

    if source is None:

        return 2, (
            f"Unsupported type: {type_}\n"
            f'Supported type: [{" ".join(IGNORE_TEMPLATE)}]. Case insensitive.'
        )

    if not (writting and path.parent.is_dir()):
        width, _ = get_terminal_size()
        return 1, "{sep}{source}{sep}".format(sep="=" * width, source=source)

    with path.open(mode="w+") as f:
        f.write(source)

    return 0, f"Write gitignore file successful. Path: {path}"



# -*- coding:utf-8 -*-

from typing import Dict, List
import os
import textwrap

from .base import ShellCompletion

_TEMPLATE_ZSH: str = """\
#compdef %(prog)s

#--------------------------------------------------------------------
# This completion script is generated automatically by parsing
# parameters.
#
#--------------------------------------------------------------------
# Author
#--------
#
# * Zachary Zhang (https://github.com/zlj-zz)
#
#--------------------------------------------------------------------

%(tools)s

%(func_name)s(){
  local curcontext="$curcontext" state line ret=1
  typeset -A opt_args

  %(arguments)s
  %(cases)s

  return ret
}

compdef %(func_name)s %(prog)s
"""

_TEMP_A: str = """
_arguments -C \\
%s
&& ret=0

"""

_TEMP_V: str = """\
__%s_values() {
  _values '' \\
%s
  && ret=0
}
"""


class ZshCompletion(ShellCompletion):

    _SHELL: str = "zsh"

    try:
        _INJECT_PATH: str = os.environ["HOME"] + "/.zshrc"
    except:
        _INJECT_PATH: str = ""

    _template_source: str = _TEMPLATE_ZSH

    def _process_arguments(self, _arguments) -> str:
        res = []

        for one in _arguments:
            names = one[0].split()
            # help msg not allow include `[]`
            help_str = one[1].replace("[", "`").replace("]", "`")

            if len(names) == 1:
                res.append(f"'{one[0]}[{help_str}]' \\")
            else:
                res.append(
                    "{%(keys)s}'[%(desc)s]' \\"
                    % {"keys": ",".join(names), "desc": help_str}
                )

        return res

    def _process_sub_commands(
        self, _sub_opts: Dict, _sub_opt_comps: List, relationship: List, idx: int = 1
    ):
        relation_str = ""

        for opt_name, opt_args in _sub_opts.items():
            _a = opt_args.get("_arguments", [])
            _a = self._process_arguments(_a)

            _s = []
            if _sub_c := opt_args.get("_sub_opts"):
                for n, x in _sub_c.items():
                    if help_str := x.get("help"):
                        # help msg not allow include `[]`
                        help_str = help_str.replace("[", "`").replace("]", "`")
                        _s.append(
                            f"'{n}[{help_str}]' \\",
                        )
                self._process_sub_commands(
                    _sub_c, _sub_opt_comps, relationship, idx + 1
                )
            if _a or _s:
                _sub_opt_comps.append(
                    _TEMP_V % (opt_name, textwrap.indent("\n".join([*_a, *_s]), "    "))
                )
                relation_str += f"    {opt_name}) __{opt_name}_values ;;\n"

        if relation_str:
            relation_str = (
                textwrap.dedent(
                    """
                    if [[ ${#line} -eq %s ]]; then
                      case $line[%s] in
                    %s
                      esac
                    fi
                    """
                )
                % (idx + 1, idx, relation_str)
            )
            relationship.append(relation_str)

    def args2complete(self, args_dict: Dict) -> str:
        prog_handle: str = args_dict["prog"]
        args: Dict = args_dict["args"]

        _arguments, _positions, _sub_opts = self._parse(args)

        # Generate zsh comp `_arguments` string.
        _arguments = self._process_arguments(_arguments)

        _sub_opt_str = ""
        _sub_sub_opt_comps = []
        _sub_relationship = []

        if _sub_opts:
            _subs = [
                f"'{opt_name}[{opt_args['help']}]' \\"
                for opt_name, opt_args in _sub_opts.items()
            ]

            _sub_opt_str = textwrap.dedent(
                """
                ######################
                # sub-commands helper
                ######################
                """
            ) + _TEMP_V % (
                "sub_opt",
                textwrap.indent("\n".join(_subs), "    "),
            )

            self._process_sub_commands(
                _sub_opts, _sub_sub_opt_comps, _sub_relationship, idx=1
            )

        _opt_case_str = ""

        if _sub_opt_str:
            _arguments.append("'1: :->opts'\\")
            _opt_case_str = textwrap.dedent(
                """
                case $state in
                opts) __sub_opt_values ;;
                """
            )

        if _sub_sub_opt_comps:
            _arguments.append("'*::arg:->args'\\")
            _opt_case_str += "  args)\n"
            _opt_case_str += textwrap.indent("\n".join(_sub_relationship), " " * 4)

        if _opt_case_str:
            _opt_case_str += "esac"

        arguments_str = _TEMP_A % textwrap.indent("\n".join(_arguments), " " * 2)

        return self._template_source % {
            "func_name": self.func_name,
            "prog": self.prog_name,
            "tools": "\n".join([*_sub_sub_opt_comps, _sub_opt_str]),
            "arguments": textwrap.indent(arguments_str, " " * 2),
            "cases": textwrap.indent(_opt_case_str, " " * 2),
        }

    def generate_resource(self) -> str:
        return self.args2complete(self.complete_vars)



# -*- coding:utf-8 -*-
from typing import TYPE_CHECKING, List, Optional, Any
import os, logging
from time import sleep

from plenty import get_console
from plenty.style import Style

from .const import IS_WIN
from .tui.event_loop import EventLoop, ExitEventLoop
from .tui.screen import Screen
from .tui.widgets import SwitchWidget, RowPanelWidget, CmdRunner, ConfirmWidget
from .gitlib.options import GitOption

if TYPE_CHECKING:
    from .gitlib.model import File, Commit, Branch


Logger = logging.getLogger(__name__)

# git option handler
git = GitOption()
console = get_console()

GREEN = "#98FB98"
RED = "#F08080"
BLUE = "#6495ED"
YELLOW = "#F0E68C"


class BranchPanel(RowPanelWidget):
    def get_raw_data(self) -> List["Branch"]:
        return git.load_branches()

    def process_raw_data(self, raw_data: List["Branch"]) -> List[str]:
        processed_branches = []
        for branch in raw_data:
            if branch.is_head:
                processed_branches.append(f"* {Style(color=GREEN).render(branch.name)}")
            else:
                processed_branches.append(f"  {branch.name}")

        return processed_branches

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print(f"{self.cursor} {line}")
        else:
            print(f"  {line}")

    def keyevent_help(self) -> str:
        return "space: switch current branch.\n"

    def process_keyevent(self, input_key: str, cursor_row: int) -> bool:
        if input_key == "q":
            raise ExitEventLoop
        elif input_key == " ":
            local_branch = self.raw_data[cursor_row - 1]
            if local_branch.is_head:
                return

            err = git.checkout_branch(local_branch.name)
            if "error" in err:
                print(Style(color=RED).render(err), sep="")
                sleep(2)
            self.emit("update")


class StatusPanel(RowPanelWidget):
    repo_path, repo_conf = git.get_repo_info()

    def get_raw_data(self) -> List["File"]:
        return git.load_status(self.size[0], icon=True)

    def process_raw_data(self, raw_data: List[Any]) -> List[str]:
        return (
            [file.display_str for file in raw_data]
            if raw_data
            else ["No status changed."]
        )

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if self.raw_data:
            if is_cursor_row:
                print(f"{self.cursor} {line}")
            else:
                print(f"  {line}")
        else:
            # No data, print tip.
            print(line)

    def keyevent_help(self) -> str:
        return (
            "a / space: toggle storage or unstorage file.\n"
            "d: discard the file changed.\n"
            "i: append the file to `.gitignore`.\n"
            "e: open file with default editor.\n"
            "↲ : check file diff.\n"
        )

    def process_keyevent(self, input_key: str, cursor_row: int) -> Optional[int]:
        if input_key in {"a", " "}:
            git.switch_file_status(self.raw_data[cursor_row - 1], path=self.repo_path)
            self.emit("update")
        elif input_key == "d":
            if ConfirmWidget("discard all changed? [y/n]:").run():
                git.discard_file(self.raw_data[cursor_row - 1], path=self.repo_path)
            self.emit("update")
            return cursor_row - 1
        elif input_key == "i":
            git.ignore_file(self.raw_data[cursor_row - 1])
            self.emit("update")
        elif input_key == "e":
            # editor = os.environ.get("EDITOR", None)
            if editor := os.environ.get("EDITOR", None):
                CmdRunner(
                    f'{editor} "{self.raw_data[cursor_row - 1].name}"',
                    path=self.repo_path,
                )

        elif input_key == "enter":
            self.widget.set_file(self.raw_data[cursor_row - 1], self.repo_path)
            self.widget.activate()
        elif input_key == "q":
            raise ExitEventLoop


class FilePanel(RowPanelWidget):
    _file = None
    _repo_path = None

    def set_file(self, file: "File", path: str):
        if self._file != file:
            self.size = None
            self.cursor_row = 1
            self._file = file
        if self._repo_path != path:
            self._repo_path = path

    def get_raw_data(self) -> List[Any]:
        return git.load_file_diff(
            self._file.name,
            self._file.tracked,
            self._file.has_staged_change,
            path=self._repo_path,
        ).split("\n")

    def process_keyevent(self, input_key: str, cursor_row: int) -> bool:
        if input_key == "q":
            self.deactivate()

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print(Style(bg_color=BLUE).render(line))
        else:
            print(line)


class CommitPanel(RowPanelWidget):
    def get_raw_data(self) -> List["Commit"]:
        branch_name = git.get_head()
        return git.load_commits(branch_name)

    def process_raw_data(self, raw_data: List[Any]) -> List[str]:
        color_data = []
        for commit in raw_data:
            c = YELLOW if commit.is_pushed() else RED
            sha = commit.sha[:7]
            msg = commit.msg
            color_data.append(Style(color=c).render(f"{sha} {msg}"))

        return color_data

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print(f"{self.cursor} {line}")
        else:
            print(f"  {line}")

    def process_keyevent(self, input_key: str, cursor_row: int) -> bool:
        if input_key == "enter":
            self.widget.set_commit(self.raw_data[cursor_row - 1])
            self.widget.activate()
        elif input_key == "q":
            raise ExitEventLoop

    def keyevent_help(self) -> str:
        return "↲ : check commit diff.\n"


class CommitStatusPanel(RowPanelWidget):
    _commit = None

    def set_commit(self, commit: "Commit"):
        if self._commit != commit:
            self.size = None
            self.cursor_row = 1
            self._commit = commit

    def get_raw_data(self) -> List[Any]:
        return git.load_commit_info(self._commit.sha).split("\n")

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print(Style(bg_color=BLUE).render(line))
        else:
            print(line)

    def process_keyevent(self, input_key: str, cursor_row: int) -> bool:
        if input_key == "q":
            self.deactivate()


class ModelSwitcher(SwitchWidget):
    def process_keyevent(self, key: str) -> Optional[int]:
        if key in "123456789":
            return int(key) - 1
        elif key in {"h"}:
            self.set_current(self.current - 1)
        elif key in {"l"}:
            self.set_current(self.current + 1)


def tui_main(index=None, help_wait=1.5):
    # tui interaction interface not support windows.
    if IS_WIN:
        console.echo("`Terminal interaction not support windows now.`<#FF0000>")
        return

    if not git.get_repo_info()[0]:
        console.echo("`Please run in a git repo dir.`<tomato>")
        return

    status = StatusPanel(widget=FilePanel(), help_wait=help_wait)
    commit = CommitPanel(widget=CommitStatusPanel(), help_wait=help_wait)
    branch = BranchPanel(help_wait=help_wait)
    switcher = ModelSwitcher(sub_widgets=[status, commit, branch])

    if index:
        start_idx = index
        switcher.set_current(int(start_idx) - 1)

    screen = Screen(switcher)
    main_loop = EventLoop(screen)
    main_loop.set_input_timeouts(0.125)
    main_loop.run()



# -*- coding:utf-8 -*-

"""This file save some model class of git info class."""

from typing import List
from dataclasses import dataclass

# flake8: noqa
@dataclass
class File:
    """Model class of git file."""

    # File path relative to Git.
    name: str
    # Display string, may has color.
    display_str: str

    # status string, like: 'MM'
    short_status: str

    # file whether has staged change.
    has_staged_change: bool

    # file whether has unstaged change.
    has_unstaged_change: bool

    # Is the file on the tracking tree.
    tracked: bool

    # file whether deleted.
    deleted: bool

    # file whether added.
    added: bool

    # file whether has merged conflict.
    has_merged_conflicts: bool

    # file whether has inline merged conflict.
    has_inline_merged_conflicts: bool


@dataclass
class Commit:
    """Model class of a git commit."""

    # The commit sha value.
    sha: str

    # The commit simple msg.
    msg: str

    # The author of the commit.
    author: str

    # The created time of the commit.
    unix_timestamp: int

    # The commit status, 'pushed' or 'unpushed'.
    status: str

    # The commit detail msg.
    extra_info: str

    # Tag list of the commit.
    tag: List

    action: str = ""

    def is_pushed(self) -> bool:
        return self.status == "pushed"


@dataclass
class Branch:
    """Model class of git branch."""

    # The branch name.
    name: str

    # This should be `True` if the branch can be pushing.
    pushables: str

    # This should be `True` if the branch can be pulling.
    pullables: str

    # This should be `True` if the branch is head.
    is_head: bool

    # The branch up-stream name. Default is "".
    upstream_name: str = ""



# -*- coding:utf-8 -*-
import pytest
from .utils import analyze_it

from pigit.entry import pigit


@pytest.mark.parametrize(
    "command",
    [
        "--report",
        "--config",  # git config
        "--information",  # git information
        "--count",  # code counter
        "--create-config",
        "cmd",
        "cmd -s",
        "cmd -t",
        "cmd -p branch",
        "cmd ws",
        "cmd -t",
        "repo ll",
    ],
)
def test_color_command(command: str):
    print()
    pigit(command.split())



import sys

PYTHON_VERSION = sys.version_info[:3]
if PYTHON_VERSION < (3, 8, 5):
    print(
        "The current version of pigit does not support less "
        "than Python3.8, more version please check https://pypi.org/project/pigit/"
    )
    exit(0)

try:
    LONG_DESCRIPTION = open("README.md", encoding="utf-8").read()
except Exception:
    LONG_DESCRIPTION = """# pigit

A terminal tool for git. When we use git, do you feel very uncomfortable with too long commands.
For example: `git status --short`, this project can help you improve it. This project is written in Python.
Now most UNIX like systems come with Python. So you can easily install and use it.

## Installation

### Pip

```bash
pip install -U pigit
```

### Source

```bash
git clone https://github.com/zlj-zz/pigit.git
cd pigit
make install
# or
python setup.py install  # On windows
```
    """

from setuptools import setup, find_packages
from pigit.const import __project__, __version__, __author__, __email__, __url__

setup(
    name=__project__,
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="Simple terminal tool of Git.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=__url__,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: System :: Shells",
        "Topic :: Software Development",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=['plenty==1.0.2'],
    entry_points="""
        [console_scripts]
        pigit=pigit.entry:main
    """,
    python_requires=">=3.8",
)



# /usr/bin/env python3
# -*- coding:utf-8 -*-

# --------------------------------------------------------
# This file is for `python -m pigit`.
# When you run a python module directly, the interpreter
# looks for the `__main__.py` file in the module.
# --------------------------------------------------------

from pigit.entry import main

if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-
from typing import Any, Callable, List, Optional, Tuple
from abc import ABC, abstractmethod
import time
from math import ceil

from plenty.str_utils import get_width
from plenty.style import Style

from .console import Term
from ..common.utils import exec_cmd, confirm


class Widget(ABC):
    _activation: bool = False

    def activate(self):
        self._activation = True

    def deactivate(self):
        self._activation = False

    def is_activation(self):
        return self._activation

    @abstractmethod
    def _process_event(self):
        """Method of processing keyboard event.

        It will be instance in sub-class.
        """

    @abstractmethod
    def _render(self, size):
        """Method of render a widget.

        It will be instance in sub-class.
        """


class SwitchWidget(Widget):
    _activation: bool = True

    def __init__(self, sub_widgets: List = None, start_idx: int = 0):
        self.idx = start_idx

        if not sub_widgets:
            self.sub_widgets: List[Widget] = []
            self.sub_widgets_count = 0
        else:
            self.sub_widgets = sub_widgets
            self.sub_widgets_count = len(sub_widgets)

    def add(self, widget: Widget):
        """Add new sub widget."""
        self.sub_widgets.append(widget)
        self.sub_widgets_count += 1

    @property
    def current(self) -> int:
        return self.idx

    def set_current(self, idx: int):
        """Set the top sub widget, if index is valid."""
        if self.idx != idx and 0 <= idx < self.sub_widgets_count:
            self.sub_widgets[self.idx].deactivate()
            self.idx = idx
            self.sub_widgets[self.idx].activate()

    @abstractmethod
    def process_keyevent(self, key: str) -> Optional[int]:
        """Costom process keyboard event, instance in sub-class."""

    def _process_event(self, key: str):
        # If the result of process is int, then switch the page.
        # Else let current sub widget to process the key.
        next_idx = self.process_keyevent(key)
        if isinstance(next_idx, int):
            self.set_current(next_idx)
        else:
            self.sub_widgets[self.idx]._process_event(key)

    def _render(self, size):
        """
        This widget cannot render any, call current sub widget ``_render``
        """

        current_sub_widget = self.sub_widgets[self.idx]
        if not current_sub_widget.is_activation():
            current_sub_widget.activate()
        current_sub_widget._render(size)


class RowPanelWidget(Widget):
    def __init__(
        self,
        cursor: Optional[str] = None,
        help_wait: float = 1.5,
        widget: Widget = None,
        **kwargs: Any,
    ) -> None:
        self.widget = widget
        self.size = None

        self.cursor = "→" if not cursor or get_width(ord(cursor)) != 1 else cursor
        self.help_wait = help_wait

        # Initialize.
        self.cursor_row: int = 1
        self.display_range: List = None  # Allow display row range.

        self.extra = 0  # Extra occupied row.

        self.raw_data: List = None
        self.show_data: List = None

        self.update_raw: bool = False

        for key, value in kwargs.items():
            setattr(self, f"_ex_{key}", value)

    @abstractmethod
    def get_raw_data(self) -> List[Any]:
        """How to get the raw data."""

    def process_raw_data(self, raw_data: List[Any]) -> List[str]:
        return raw_data

    def generate_show_data(
        self, raw_data: List[str], width: int
    ) -> List[Tuple[str, int]]:
        """
        Process the raw data, and indicate the number of additional
        rows that need to be occupied when each piece of data is displayed.
        """

        new_list = []
        for line in raw_data:
            text = Style.plain(line)
            count = sum(get_width(ord(ch)) for ch in text)
            # [float] is to solve the division of python2 without
            # retaining decimal places.
            new_list.append((line, ceil(count / width) - 1))
        return new_list

    @abstractmethod
    def print_line(self, line: str, is_cursor_row: bool) -> None:
        """How to output one line.

        May has some different when current line is cursor line.
        Support to process cursor line specially.
        """

    def update(self):
        self.display_range = [1, self.size[1] - 1]
        self.raw_data: List[Any] = self.get_raw_data()
        self.show_data = self.generate_show_data(
            self.process_raw_data(self.raw_data), self.size[0]
        )

    def emit(self, name: str, cb: Optional[Callable] = None):
        if name == "update":
            self.update()

    def _render(self, size):
        if self.widget and self.widget.is_activation():
            self.widget._render(self.size)
            return
        if self.size != size:
            self.size = size
            self.update()

        # Adjust display row range.
        while self.cursor_row < self.display_range[0]:
            self.display_range = [i - 1 for i in self.display_range]
        while self.cursor_row + self.extra > self.display_range[1]:
            self.display_range = [i + 1 for i in self.display_range]

        # Every time refresh the output, need to recalculate the
        # number of additional rows, so need to reset to zero.
        self.extra: int = 0

        # Print needed display part.
        for index, item in enumerate(self.show_data, start=1):
            if self.display_range[0] <= index <= self.display_range[1] - self.extra:
                line, each_extra = item
                self.print_line(line, index == self.cursor_row)
                self.extra += each_extra

    def _process_event(self, key: str):
        if self.widget and self.widget.is_activation():
            # If has sub widget and it's activation.
            self.widget._process_event(key)
        elif self.is_activation():
            # Process key.
            if key in {"j", "down"}:
                # select pre file.
                self.cursor_row += 1
                self.cursor_row = min(self.cursor_row, len(self.show_data))

            elif key in {"k", "up"}:
                # select next file.
                self.cursor_row -= 1
                self.cursor_row = max(self.cursor_row, 1)

            elif key in {"J"}:
                # scroll down 5 lines.
                self.cursor_row += 5
                self.cursor_row = min(self.cursor_row, len(self.show_data))

            elif key in {"K"}:
                # scroll up 5 line
                self.cursor_row -= 5
                self.cursor_row = max(self.cursor_row, 1)

            elif key in {"?", "H"}:
                print(Term.clear_screen)
                print(
                    (
                        "k / ↑: select previous line.\n"
                        "j / ↓: select next line.\n"
                        "J: Scroll down 5 lines.\n"
                        "K: Scroll down 5 lines.\n"
                        "? / H : show help, wait {}s and exit.\n" + self.keyevent_help()
                    ).format(self.help_wait)
                )
                time.sleep(self.help_wait)

            else:
                #
                next_cursor = self.process_keyevent(key, self.cursor_row)
                if next_cursor is not None and 1 <= next_cursor <= len(self.show_data):
                    self.cursor_row = next_cursor

    @abstractmethod
    def process_keyevent(self, input_key: str, cursor_row: int) -> Optional[int]:
        """Handles keyboard events other than movement.

        Args:
            input_key (str): keyboard string.
            cursor_row (int): current line.
            data (Any): raw data.

        Returns:
            int: next cursor index.
        """

    def keyevent_help(self) -> str:
        """Get extra keyevent help message.

        Returns:
            str: help message string.
        """

        return ""


class ConfirmWidget:
    def __init__(self, msg: str, default: bool = True) -> None:
        self.msg = msg
        self.default = default

    def run(self):
        print(Term.clear_screen, end="")
        return confirm(self.msg, self.default)


class CmdRunner:
    def __init__(self, cmd: str, auto_run: bool = True, path: str = ".") -> None:
        self.cmd = cmd
        self.auto_run = auto_run
        self.run_path = path

        if self.auto_run:
            self.run()

    def run(self):
        print(Term.normal_screen)
        res_code = exec_cmd(self.cmd, cwd=self.run_path, reply=False)
        print(Term.alt_screen)
        return res_code



import pytest, os, copy
from pprint import pprint
from .utils import analyze_it
from .conftest import _PIGIT_PATH

from pigit.cmdparse.parser import command, Parser
from pigit.cmdparse.shellcompletion.base import ShellCompletion
from pigit.cmdparse.shellcompletion import (
    ZshCompletion,
    BashCompletion,
    FishCompletion,
    shell_complete,
    get_shell,
)
from pigit.gitlib.shortcmds import GIT_CMDS

argparse_dict = {
    "prog": "pigit",
    "prefix_chars": "-",
    "description": "Pigit TUI is called automatically if no parameters are followed.",
    "args": {
        "-v --version": {
            "action": "version",
            "help": "Show version and exit.",
            "version": "Version: ",
        },
        "-d --debug": {
            "action": "store_true",
            "help": "Current runtime in debug mode.",
        },
        "--out-log": {"action": "store_true", "help": "Print log to console."},
        "tools": {
            "type": "groups",
            "title": "tools arguments",
            "description": "Auxiliary type commands.",
            "args": {
                "-c --count": {
                    "nargs": "?",
                    "const": ".",
                    "type": str,
                    "metavar": "PATH",
                    "help": "Count the number of codes and output them in tabular form."
                    "A given path can be accepted, and the default is the current directory.",
                },
                "--create-config": {
                    "action": "store_true",
                    "help": "Create a pre-configured file of PIGIT."
                    "(If a profile exists, the values available in it are used)",
                },
            },
        },
        "cmd": {
            "type": "sub",
            "help": "git short command.",
            "description": "If you want to use some original git commands, please use -- to indicate.",
            "args": {
                "command": {
                    "nargs": "?",
                    "type": str,
                    "default": None,
                    "help": "Short git command or other.",
                },
                "args": {
                    "nargs": "*",
                    "type": str,
                    "help": "Command parameter list.",
                },
                "-t --types": {
                    "action": "store_true",
                    "help": "List all command types and exit.",
                },
                "set_defaults": {"func": range},
            },
        },
    },
}


def test_generate_about_dict():
    parser = Parser.from_dict(argparse_dict)
    parser.print_help()

    ShellCompletion._SHELL = ""
    ShellCompletion._INJECT_PATH = ""
    args = parser.to_dict()["args"]
    pprint(ShellCompletion("", {})._parse(args))


class TestCompletion:
    real = True
    prog = "pigit"
    script_dir = os.path.join(_PIGIT_PATH, "docs")

    if real:
        from pigit.entry import pigit

        complete_vars = pigit.to_dict()
        complete_vars["args"]["cmd"]["args"].update(
            {k: {"help": v["help"], "args": {}} for k, v in GIT_CMDS.items()}
        )
    else:
        complete_vars = copy.deepcopy(argparse_dict)
        cmd_temp = complete_vars["args"]["cmd"]["args"]
        cmd_temp.update(
            {k: {"help": v["help"], "args": {}} for k, v in GIT_CMDS.items()}
        )

    def test_error_complete_vars(self):
        with pytest.raises(TypeError):
            BashCompletion("test", "xxx", ".")

    def print(self, c):
        print(c.prog_name)
        print(c.script_name)
        print(c.inject_path)

        source = c.generate_resource()
        print(source)

        c.write_completion(source)

    @analyze_it
    def test_bash(self):
        c = BashCompletion(None, self.complete_vars, self.script_dir)
        self.print(c)

    @analyze_it
    def test_zsh(self):
        c = ZshCompletion("pigit-dev", self.complete_vars, self.script_dir)
        self.print(c)

    def test_fish(self):
        c = FishCompletion(self.prog, self.complete_vars, self.script_dir)
        self.print(c)

    def test_get_shell(self):
        print(get_shell())

    def test_action(self):
        shell_complete(self.complete_vars, "bash", "xxx", ".", None, "./test.txt")



# -*- coding:utf-8 -*-
# The PIGIT terminal tool entry file.

from typing import Dict, List
import os, logging, textwrap

from plenty import get_console

from .log import setup_logging
from .config import Config
from .cmdparse.parser import command, argument, Namespace
from .const import (
    VERSION,
    REPOS_PATH,
    PIGIT_HOME,
    LOG_FILE_PATH,
    EXTRA_CMD_MODULE_NAME,
    EXTRA_CMD_MODULE_PATH,
    CONFIG_FILE_PATH,
    COUNTER_DIR_PATH,
    IS_FIRST_RUN,
)
from .common.utils import confirm
from .common.func import dynamic_default_attrs, time_it
from .gitlib.processor import ShortGitter, GIT_CMDS, get_extra_cmds
from .gitlib.options import GitOption
from .gitlib.ignore import create_gitignore
from .info import introduce, GitConfig


Logger = logging.getLogger(__name__)


# ===============
# Configuration.
# ===============
CONFIG = Config(
    path=CONFIG_FILE_PATH, version=VERSION, auto_load=True
).output_warnings()

setup_logging(
    debug=CONFIG.debug_open,
    log_file=None if CONFIG.log_output else LOG_FILE_PATH,
)


# ==============
# Global handle
# ==============
git = GitOption(repo_json_path=REPOS_PATH)
console = get_console()


# =====================
# main command `pigit`
# yapf: disable
# =====================
@command("pigit", description="Pigit TUI is called automatically if no parameters are followed.")
@argument("-r --report", action="store_true", help="Report the pigit desc and exit.")
@argument("-f --config", action="store_true", help="Display the config of current git repository and exit.")
@argument("-i --information", action="store_true", help="Show some information about the current git repository.")
def pigit(args: Namespace, _) -> None:
    if args.report:
        console.echo(introduce())

    elif args.create_config:
        return CONFIG.create_config_template()

    elif args.config:
        console.echo(GitConfig(format_type=CONFIG.git_config_format).generate())

    elif args.information:
        console.echo(git.get_repo_desc(include_part=CONFIG.repo_info_include))

    elif args.complete:
        # Generate competion vars dict.
        complete_vars = pigit.to_dict()
        complete_vars["args"]["cmd"]["args"].update(
            {k: {"help": v["help"], "args": {}} for k, v in GIT_CMDS.items()}
        )

        from .cmdparse.shellcompletion import shell_complete

        shell_complete(complete_vars, PIGIT_HOME, inject=True)
        return None

    elif args.ignore_type:
        _, msg = create_gitignore(args.ignore_type,writting=True)
        console.echo(msg)

    elif args.count:
        from .codecounter import CodeCounter

        path = os.path.abspath(args.count) if args.count != "." else os.getcwd()
        CodeCounter(
            count_path=path,
            use_ignore=CONFIG.counter_use_gitignore,
            result_saved_path=COUNTER_DIR_PATH,
            format_type=CONFIG.counter_format,
            use_icon=CONFIG.counter_show_icon,
        ).run(
            show_invalid=CONFIG.counter_show_invalid,
        )
        return None

    # Don't have invalid command list.
    # if not list(filter(lambda x: x, vars(known_args).values())):
    else:
        from .interaction import tui_main

        if IS_FIRST_RUN:
            introduce()
            if not confirm("Input `enter` to continue:"):
                return

        tui_main(help_wait=CONFIG.tui_help_showtime)


pigit.add_argument("-v", "--version", action="version", help="Show version and exit.", version=f"Version:{VERSION}")

tools_group = pigit.add_argument_group(title="tools arguments", description="Auxiliary type commands.")
tools_group.add_argument("-c", "--count", nargs="?", const=".", type=str, metavar="PATH",
    help="""Count the number of codes and output them in tabular form.
    A given path can be accepted, and the default is the current directory.
    """,)
tools_group.add_argument( "--create-ignore", type=str, metavar="TYPE", dest="ignore_type",
    help="""Create a demo .gitignore file. Need one argument, the type of gitignore.""")
tools_group.add_argument("-C", "--complete", action="store_true",
    help="Add shell prompt script and exit. (Supported bash, zsh, fish)")
tools_group.add_argument("--create-config", action="store_true",
    help="Create a pre-configured file of PIGIT."
    "(If a profile exists, the values available in it are used)",)

# =============================================
# sub command `cmd`
# =============================================
@pigit.sub_parser("cmd", help="git short command.")
@argument("--shell", action="store_true", help="Go to the pigit shell mode.")
@argument("-t --types", action="store_true", help="List all command types and exit.")
@argument("-p --show-part-command", type=str, metavar="TYPE", dest="command_type",
    help='According to given type to list available short command and wealth and exit.')
@argument("-s --show-commands", action="store_true", help="List all available short command and wealth and exit.")
@argument("args", nargs="*", type=str, help="Command parameter list.")
@argument("command", nargs="?", type=str, default=None, help="Short git command or other.")
def _cmd_func(args: Namespace, unknown: List):
    """If you want to use some original git commands, please use -- to indicate."""

    # If you want to manipulate the current folder with git,
    # try adding it to repos automatically.
    if CONFIG.repo_auto_append:
        repo_path, repo_conf = git.get_repo_info()
        git.add_repos([repo_path])

    # Init extra custom cmds.
    extra_cmd: Dict = {
        # "shell": {
        #     "command": lambda _: shell_mode(git_processor),
        #     "type": "func",
        #     "help": "Into PIGIT shell mode.",
        # },  # only for tips.
    }
    extra_cmd.update(get_extra_cmds(EXTRA_CMD_MODULE_NAME, EXTRA_CMD_MODULE_PATH))

    git_processor = ShortGitter(
        extra_cmds=extra_cmd,
        command_prompt=CONFIG.cmd_recommend,
        show_original=CONFIG.cmd_show_original,
    )

    if args.shell:
        from .shellmode import PigitShell
        shell_ = PigitShell(git_processor)
        # print(shell_.__dir__())
        shell_.cmdloop()
        return

    if args.show_commands:
        return git_processor.print_help()

    if args.command_type:
        return git_processor.print_help_by_type(args.command_type)

    if args.types:
        return git_processor.print_types()

    if args.command:
        command = args.command
        args.args.extend(unknown)
        git_processor.do(command, args.args)
        return None
    else:
        console.echo("`pigit cmd -h`<ok> for help.")


# =============================================
# sub command `repo`
# =============================================
repo = pigit.sub_parser("repo", help="repos options.")(lambda _, __: print("-h help"))


@repo.sub_parser("add", help="add repo(s).")
@argument("--dry-run", action="store_true", help="dry run.")
@argument("paths", nargs="+", help="path of reps(s).")
def repo_add(args, _):
    if added := git.add_repos(args.paths, args.dry_run):
        console.echo(f"Found {len(added)} new repo(s).")
        for path in added:
            console.echo(f"\t`{path}`<sky_blue>")
    else:
        console.echo("`No new repos found!`<tomato>")


@repo.sub_parser("rm", help="remove repo(s).")
@argument("--path", action="store_true", help="remove follow path, defult is name.")
@argument("repos", nargs="+", help="name or path of repo(s).")
def repo_rm(args, _):
    res = git.rm_repos(args.repos, args.path)
    for one in res:
        console.echo(f"Deleted repo. name: '{one[0]}', path: {one[1]}")


@repo.sub_parser("rename", help="rename a repo.")
@argument("new_name", help="the new name of repo.")
@argument("repo", help="the name of repo.")
def repo_rename(args, _):
    success, msg = git.rename_repo(args.repo, args.new_name)
    console.echo(msg)


@repo.sub_parser("ll", help="display summary of all repos.")
@argument("--simple", action="store_true", help="display simple summary.")
@argument("--revese", action="store_true", help="revese to display invalid repo.")
def repo_ll(args, _):
    simple = args.simple
    revese = args.revese

    for info in git.ll_repos(revese=revese):
        if simple:
            if revese:
                console.echo(f"{info[0][0]:<20} {info[1][1]:<15}")
            else:
                console.echo(f"{info[0][0]:<20} {info[1][1]:<15} {info[5][1]}")
        else:
            if revese:
                summary_string = textwrap.dedent(
                    f"""\
                    b`{info[0][0]}`
                        {info[1][0]}: `{info[1][1]}`<sky_blue>
                    """
                )
            else:
                summary_string = textwrap.dedent(
                    f"""\
                    b`{info[0][0]}`
                        {info[1][0]}: {info[1][1]}
                        {info[2][0]}: {info[2][1]}
                        {info[3][0]}: `{info[3][1]}`<khaki>
                        {info[4][0]}: `{info[4][1]}`<ok>
                        {info[5][0]}: `{info[5][1]}`<sky_blue>
                    """
                )
            console.echo(summary_string)


repo.sub_parser("clear", help="clear the all repos.")(lambda _, __: git.clear_repos())


repo_options = {
    "fetch": {"cmd": "git fetch", "allow_all": True, "help": "fetch remote update"},
    "pull": {"cmd": "git pull", "allow_all": True, "help": "pull remote updates"},
    "push": {"cmd": "git push", "allow_all": True, "help": "push the local updates"},
}
for subcmd, prop in repo_options.items():
    help_string = f'{h.strip()} for repo(s).' if (h:=prop.get("help")) else "NULL"
    repo.sub_parser(subcmd, help=help_string)(
        argument("repos", nargs="*", help="name of repo(s).")(
            dynamic_default_attrs(
                lambda args, _, cmd: git.process_repo_option(args.repos, cmd),
                cmd=prop['cmd']
            )
        )
    )


# =============================================
# sub command `open`
# =============================================
@pigit.sub_parser("open", help="open remote repository in web browser.")
@argument("-p --print", action="store_true", help="only print the url at the terminal, but do not open it.")
@argument("-c --commit", help="the current commit in the repo website.")
@argument("-i --issue", help="the given issue of the repository.")
@argument("branch", nargs="?", default=None, help="the branch of repository.")
def _open_func(args: Namespace, _):
    code, msg = git.open_repo_in_browser(
        branch=args.branch, issue=args.issue, commit=args.commit, print=args.print
    )
    console.echo(msg)


# =============================================
# terminal entry
# =============================================
@time_it
def main():
    try:
        pigit()
    except (KeyboardInterrupt, EOFError):
        raise SystemExit(0) from None



from typing import Dict, Literal, Optional, Tuple
import os, re

from plenty.table import UintTable
from plenty import box

from .const import __version__, __url__
from .gitlib.options import GitOption


git = GitOption()


def introduce() -> str:
    """Print the description information."""

    # Print version.
    introduce_str = """\
` ____ ___ ____ ___ _____
|  _ \\_ _/ ___|_ _|_   _|
| |_) | | |  _ | |  | |
|  __/| | |_| || |  | |
|_|  |___\\____|___| |_|`<pink> version: {version}

{git_version}

b`Local path`: u`{local_path}`<sky_blue>

b`Description:`
  Terminal tool, help you use git more simple. Support Linux, MacOS and Windows.
  The open source path on github: u`{url}`<sky_blue>

You can use `-h`<ok> or `--help`<ok> to get help and usage.
"""

    # Print git version.
    git_version = git.git_version or "`Don't found Git, maybe need install.`<error>"

    return introduce_str.format(
        version=__version__,
        git_version=git_version,
        local_path=os.path.dirname(__file__.replace("./", "")),
        url=__url__,
    )


# ============
# Config info
# ============
FormatType = Literal["normal", "table"]


class GitConfig:
    def __init__(
        self,
        repo_path: Optional[str] = None,
        format_type: FormatType = "table",
        color: bool = True,
    ) -> None:
        self.repo_path = repo_path or "."
        self.format_type = format_type
        self.color = color

    @property
    def repo_info(self) -> Tuple[str, str]:
        return git.get_repo_info(self.repo_path)

    def parse_git_config(self, config_context: str) -> Dict:
        """Retrun a dict from parsing git local config.

        Args:
            conf (str): git local config string.

        Returns:
            dict: git local config dict.
        """
        conf_list = re.split(r"\r\n|\r|\n", config_context)
        config_dict: dict[str, dict[str, str]] = {}
        config_type: str = ""

        for line in conf_list:
            line = line.strip()

            if not line:
                continue

            elif line.startswith("["):
                config_type = line[1:-1].strip()
                config_dict[config_type] = {}

            elif "=" in line:
                key, value = line.split("=", 1)
                config_dict[config_type][key.strip()] = value.strip()

            else:
                continue

        return config_dict

    def normal_config(self, config_dict: Dict[str, Dict]) -> str:
        gen = []
        color = self.color

        for tit, kv in config_dict.items():
            gen.append(f"`[{tit}]`<tomato>" if color else f"[{tit}]")
            gen.extend(
                f"\t`{k}`<sky_blue>=`{v}`<medium_violet_red>" if color else f"\t{k}={v}"
                for k, v in kv.items()
            )

        return "\n".join(gen)

    def table_config(self, config_dict: Dict[str, Dict]) -> str:
        style = ["", "pale_green" if self.color else ""]
        tb = UintTable(title="Git Local Config", box=box.DOUBLE_EDGE)

        for header, values in config_dict.items():
            tb.add_unit(header, header_style="bold", values=values, values_style=style)

        return tb

    def generate(self) -> str:
        repo_path, config_path = self.repo_info

        if not repo_path:
            return (
                "`This directory is not a git repository yet.`<error>"
                if self.color
                else "This directory is not a git repository yet."
            )

        try:
            with open(f"{config_path}/config", "r") as cf:
                context = cf.read()
        except Exception as e:
            return (
                f"`Error reading configuration file; {e}.`<error>"
                if self.color
                else f"Error reading configuration file; {e}."
            )
        else:
            config_dict = self.parse_git_config(context)
            ft = self.format_type

            if ft == "table":
                return self.table_config(config_dict)
            else:
                # Unrecognized types and normal are considered normal.
                return self.normal_config(config_dict)
